/*
 * Copyright 2017-2020, Strimzi authors.
 * License: Apache License 2.0 (see the file LICENSE or http://apache.org/licenses/LICENSE-2.0.html).
 */
package io.strimzi.kafka.oauth.services;

import io.strimzi.kafka.oauth.common.BearerTokenWithPayload;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.concurrent.ExecutorService;
import java.util.function.Consumer;
import java.util.function.Predicate;

/**
 * Sessions entries should automatically get cleared as KafkaPrincipals for the sessions get garbage collected by JVM.
 * The size of `activeSessions` at any moment in time should be about the number of currently active sessions.
 */
public class Sessions {

    private static final Object NONE = new Object();

    /**
     * The map of all 'sessions' in the form of BearerTokenWithPayload - the custom extension of OAuthBearerToken object
     * produced by our Validators at authentication time, which to us represents a session context.
     * The WeakHashMap is used to not prevent JVM from garbage collecting the completed and otherwise terminated sessions,
     * while we're still able to iterate over the active sessions.
     */
    private final Map<BearerTokenWithPayload, Object> activeSessions = Collections.synchronizedMap(new WeakHashMap<>());

    /**
     * Put a new token object into sessions cache
     *
     * @param token A token to add
     */
    public void put(BearerTokenWithPayload token) {
        activeSessions.put(token, NONE);
    }

    /**
     * Remove a token from sessions cache
     *
     * @param token A token to remove
     */
    public void remove(BearerTokenWithPayload token) {
        activeSessions.remove(token);
    }

    /**
     * Iterate over all active sessions (represented by stored token objects) applying a filter and submit a task to the passed executor for each passing token.
     * Return a list of {@link SessionFuture} instances.
     *
     * @param executor An Executor to use for submitting tasks
     * @param filter A filter to decide if the task should be submitted for specific token
     * @param task Logic to run on the token
     * @return A list of SessionFuture instances
     */
    public List<SessionFuture<?>> executeTask(ExecutorService executor, Predicate<BearerTokenWithPayload> filter,
                                           Consumer<BearerTokenWithPayload> task) {
        cleanupExpired();

        // In order to prevent the possible ConcurrentModificationException in the middle of using an iterator
        // we first make a local copy, then iterate over the copy
        ArrayList<BearerTokenWithPayload> values = new ArrayList<>(activeSessions.keySet());

        List<SessionFuture<?>> results = new ArrayList<>(values.size());
        for (BearerTokenWithPayload token: values) {
            if (filter.test(token)) {
                SessionFuture<?> current = new SessionFuture<>(token, executor.submit(() -> task.accept(token)));
                results.add(current);
            }
        }

        return results;
    }

    /**
     * Cleanup the sessions whose lifetime has expired
     */
    public void cleanupExpired() {
        // In order to prevent the possible ConcurrentModificationException in the middle of using an iterator
        // we first make a local copy, then iterate over the copy
        ArrayList<BearerTokenWithPayload> values = new ArrayList<>(activeSessions.keySet());

        long now = System.currentTimeMillis();

        // Remove expired
        for (BearerTokenWithPayload token: values) {
            if (token.lifetimeMs() <= now) {
                activeSessions.remove(token);
            }
        }
    }

    /**
     * Iterate over sessions, and find the first element matching the filter.
     *
     * @param filter A filter to apply
     * @return The first matching session
     */
    public BearerTokenWithPayload findFirst(Predicate<BearerTokenWithPayload> filter) {
        // In order to prevent the possible ConcurrentModificationException in the middle of using an iterator
        // we first make a local copy, then iterate over the copy
        ArrayList<BearerTokenWithPayload> values = new ArrayList<>(activeSessions.keySet());

        for (BearerTokenWithPayload token: values) {
            if (filter.test(token)) {
                return token;
            }
        }
        return null;
    }
}
